/*
*	This was written by Arun Sagar and is placed into the public domain.
*	The author hereby disclaims copyright to this source code.
*
*	Pay it forward. 
*/

class cReversiBoard
{
protected:
	typedef uint8 CaptureFlag;
	typedef uint8 Base3TableEntry;

	typedef struct 
	{
		uint8 left:4;
		uint8 right:4;
	} CaptureData;

	typedef CaptureData CaptureArray[8];

protected:
	static CaptureArray grgCaptureArray[];
	static CaptureFlag grgCaptureFlags[];
	static Base3TableEntry grgBase3Table[];
	static ulong grgPieceNormalizerMask[];
	static ulong grgBitCount[];

public:
//	typedef ulong Piece;
	typedef ulong Position;
	typedef enum
	{
		kPieceNone=0,
		kPieceEdge,
		kPiece1,
		kPiece2
	} Piece;
	
	typedef Position CaptureListData[2];
	typedef ushort ShortLine;
	typedef ulong LongLine;

protected:
	enum
	{
		kPieceBits=2,
		kAllEdges = AFilledInteger8(kPieceEdge,kPieceBits),
		kPieceNormalizerMask = ~AFilledInteger8(kPiece1,kPieceBits)&0xffff,
		kAllPiece1 = AFilledInteger8(kPiece1,kPieceBits),
		kAllPiece2 = AFilledInteger8(kPiece2,kPieceBits)
	};

	enum
	{
		kInfinity = LONG_MAX/2
	};

public:
	LongLine Board[8];

private:
	void StuffColumn(LongLine rLine, Position column);
	void StuffXY(LongLine rLine, Position xy);
	void StuffYX(LongLine rLine, Position yx);

protected:
	void Clear(void);
	inline static LongLine CaptureLine(LongLine rLine, Position position);
	inline static ulong CountLine(ulong) ;
	inline static CaptureData GetCaptureData(LongLine rl, Position position) {return ((grgCaptureArray[(((ushort)grgBase3Table[(Base3TableEntry)((rl)>>8)] * 81))+grgBase3Table[(Base3TableEntry)(rl)]])[position]);}
	inline static CaptureFlag IsCapturePosition(LongLine rl, Position position){return ((cReversiBoard::grgCaptureFlags[rl]<<(position)) & 0x80);}
	inline static bool IsLineFull(LongLine rL) {return 0==(kAllEdges&(~rL & (~rL>>1)));}
	inline static bool IsLineEmpty(LongLine rL) {return 0 ==(rL&kAllPiece1);}
	inline static LongLine InvertPieces(LongLine rLine){return (rLine&0xaaaa)|(((rLine>>1)^rLine)&0x5555);}

public:
	void Reset(void);

	inline static Piece OtherPiece(Piece p) {return (Piece)(1 ^ p);}

	Piece GetPiece(Position x, Position y) const {return (Piece)((this->Board[y] >> ((7-x)*kPieceBits)) & ABuildOverlayMask(kPieceBits,0));}
	inline bool IsPiece(Piece p, Position x, Position y) const {return p == (Piece)((this->Board[y] >> ((7-x)*kPieceBits)) & ABuildOverlayMask(kPieceBits,0));}
	ulong CountPieces(Piece = kPieceNone) const ;
	inline LongLine ExtractY(Position) const;
	LongLine ExtractXY(Position) const;
	LongLine ExtractYX(Position) const;

	long FillCaptureList(Piece, Position, Position, CaptureListData[]);
	
	void MakeMove(Piece p, Position x, Position y);
	bool IsMoveAvailable(Piece p) const;
	bool IsLegalMove(Piece p, Position x, Position y) const;
};


inline cReversiBoard::LongLine cReversiBoard::ExtractY(Position i) const
{
#define VEXTR(j) ((this->Board[j] & ABuildOverlayMask(kPieceBits, (7-(i))*kPieceBits)) << ((7-(j))*kPieceBits))
	return (VEXTR(0) | VEXTR(1)| VEXTR(2)| VEXTR(3)| VEXTR(4)| VEXTR(5)| VEXTR(6)|VEXTR(7)) >> (kPieceBits*(7-(i)));
#undef VEXTR
}

#define cReversiBoardTransferPiece(rlSrc, rlDst, pos) ATransferBits(rlSrc, rlDst, (pos)*kPieceBits, kPieceBits)



inline cReversiBoard::LongLine cReversiBoard::
CaptureLine(LongLine rLine, Position position)
{
#define kPCM(l) ABuildClearMask((l)*kPieceBits, (8-l)*kPieceBits)
	const static ulong rgPieceClearMask[]=
	{
		kPCM(0),kPCM(1),kPCM(2),kPCM(3),kPCM(4),kPCM(5),kPCM(6),kPCM(7)
	};
#undef kPCM
#define kPPM(n) AFilledInteger##n(kPiece1, kPieceBits) << ((8-n)*kPieceBits)
	const static ulong rgPiecePlacementMask[]=
	{
		kPPM(0), kPPM(1), kPPM(2), kPPM(3), kPPM(4), kPPM(5), kPPM(6), kPPM(7)
	};
#undef kPPM
	CaptureData rcdata = cReversiBoard::GetCaptureData(rLine, position);

	return ((((((rLine << ((position-rcdata.left)*kPieceBits)) & rgPieceClearMask[rcdata.left])
			| rgPiecePlacementMask[rcdata.left]) << (rcdata.left+1)*kPieceBits)
				& rgPieceClearMask[rcdata.right])
			| rgPiecePlacementMask[rcdata.right]) >> (position+1)*kPieceBits;
}
